package main

import (
	"context"
	"flag"
	"fmt"
	"net"
	"net/http"
	"os"
	"strings"
	"time"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/common/api/health"
	"gitlab.mashplace.com/mashplace/backend/common/generic/addr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	logger "gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/metrics"
	"gitlab.mashplace.com/mashplace/backend/common/generic/tracer"
	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/datagovuk"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/doogalcouk"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/postcodedistricts"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/tourpedia"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/uktowns"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model/postgres"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/service"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/sitemap"
	"gitlab.mashplace.com/mashplace/backend/geo/pkg/mapbox"
	"gitlab.mashplace.com/mashplace/backend/geo/version"
	grabdata "gitlab.mashplace.com/mashplace/backend/grabdata/api"
)

var envImportMinDuration = env.ResolveDuration("IMPORT_MIN_DURATION", time.Hour*24)
var envMapBoxAPIKey = env.ResolveString("MAPBOX_API_KEY", "")
var envSitemapURL = env.ResolveString("SITEMAP_URL", "https://staging.env.mashroom.com")
var (
	grabDataServiceURL = env.ResolveString("GRABDATA_SERVICE_URL", "mashroom-grabdata:10000")
	grabDataService    = grabdata.NewGrabDataServiceClient(grpc.Dial(grabDataServiceURL, "Grabdata"))
)

func main() {
	imponly := flag.Bool("importonly", false, "Just import with specified minimum duration")
	flag.Parse()

	envSitemapURL = strings.TrimSuffix(envSitemapURL, "/")

	ctx := context.Background()

	defer tracer.Close()

	conn := fmt.Sprintf(
		"postgres://%s:%s@%s:%s/%s?sslmode=%s",
		os.Getenv("POSTGRES_USERNAME"),
		os.Getenv("POSTGRES_PASSWORD"),
		os.Getenv("POSTGRES_HOST"),
		os.Getenv("POSTGRES_PORT"),
		os.Getenv("POSTGRES_DATABASE"),
		os.Getenv("POSTGRES_SSL_MODE"),
	)

	db := sqlx.MustOpen("postgres", conn)

	repo, err := postgres.NewRepository(ctx, db)
	if err != nil {
		panic(err)
	}

	httpClient := &http.Client{}

	mapboxClient, err := mapbox.NewClient(&mapbox.Config{
		GeocodingURL:       mapbox.DefaultGeocodingURL,
		IsochroneURL:       mapbox.DefaultIsochroneURL,
		Key:                envMapBoxAPIKey,
		POITypes:           mapbox.DefaultPOITypes,
		MaxContours:        4,
		MaxIsochronePeriod: time.Hour,
		HTTPClient:         httpClient,
	})
	if err != nil {
		panic(err)
	}

	sitemapGenerator, err := sitemap.NewGenerator(&sitemap.Config{
		PublicPagesURL:   "https://wordpress.mashroom.com/wp-json/wp/v2/pages?per_page=100",
		ArticlesPagesURL: "https://wordpress.mashroom.com/wp-json/wp/v2/posts",
		SitemapURL:       envSitemapURL,
		Storage:          "storage/sitemaps",
		Repo:             repo,
		HTTPClient:       httpClient,
		GrabdataCli:      grabDataService,
	})
	if err != nil {
		panic(err)
	}

	srv, err := service.NewService(&service.Config{
		Repository:       repo,
		MapboxClient:     mapboxClient,
		SitemapGenerator: sitemapGenerator,
		SitemapCountries: []*sitemap.Country{
			{
				Name: "England",
				File: "sitemapengland",
			},
			{
				Name: "Scotland",
				File: "sitemapscotland",
			},
			{
				Name: "Wales",
				File: "sitemapwales",
			},
			{
				Name: "Northern Ireland",
				File: "sitemapnorthernireland",
			},
			{
				Name: "Isle of Man",
				File: "sitemapisleofman",
			},
		},
		Importers: []importer.Importer{
			uktowns.New(repo),
			postcodedistricts.New(repo),
			tourpedia.NewLondonPOI(repo, httpClient),
			doogalcouk.NewLondonStations(repo, httpClient),
			doogalcouk.NewUkStations(repo, httpClient),
			datagovuk.NewClinics(repo, httpClient),
			datagovuk.NewHospitals(repo, httpClient),
			datagovuk.NewNaptan(repo, httpClient),
		},
		MinImportPeriod: envImportMinDuration,
	})
	if err != nil {
		panic(err)
	}

	go func() {
		_, _ = srv.UpdateDB(ctx, &geo.UpdateDBRequest{})

		_, err = srv.GenerateSitemaps(ctx, &geo.GenerateSitemapsRequest{})
		logger.WithContext(ctx).Infof("done updating sitemaps, error: %v", err)

		if imponly != nil && *imponly {
			os.Exit(0)
		}
	}()

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("geo")
	if err != nil {
		panic(err)
	}

	geo.RegisterGeoServiceServer(grpcServer, srv)

	logger.WithContext(ctx).Infoln("geo started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.WithContext(ctx).Errorln("gRPC error", err)
	}
}
