// Package datagovuk provides importers from media.nhschoices.nhs.uk
package datagovuk

import (
	"context"
	"fmt"
	"net/http"
	"strconv"
	"time"

	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/util"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type importerClinicsImpl struct {
	repository model.Repository
	client     *http.Client
}

// NewClinics returns new importer instance
func NewClinics(repo model.Repository, client *http.Client) importer.Importer {
	return &importerClinicsImpl{
		repository: repo,
		client:     client,
	}
}

func (imp *importerClinicsImpl) ID() string {
	return "mashroom-data-gov-uk-clinics"
}

func (imp *importerClinicsImpl) Source() string {
	return "http://media.nhschoices.nhs.uk/data/foi/Clinics.csv"
}

func (imp *importerClinicsImpl) Import(ctx context.Context) (success, failure int, err error) {
	resp, err := imp.client.Get(imp.Source())
	if err != nil {
		return
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	return util.ImportSplit(resp.Body, "\xAC", func(record []string) (err error) {
		if len(record) < 19 {
			return util.ErrTooShort
		}

		orgtype := record[2]
		name := record[6]
		address1 := record[7]
		address2 := record[8]
		address3 := record[9]
		postcode := record[12]
		latitude := record[13]
		longitude := record[14]
		phone := record[15]
		email := record[16]
		website := record[17]
		fax := record[18]

		lon, err := strconv.ParseFloat(longitude, 64)
		if err != nil {
			return err
		}

		lat, err := strconv.ParseFloat(latitude, 64)
		if err != nil {
			return err
		}

		address := ""

		if len(address1) > 0 {
			address += address1
		}

		if len(address2) > 0 {
			if len(address) > 0 {
				address += " "
			}

			address += address2
		}

		if len(address3) > 0 {
			if len(address) > 0 {
				address += " "
			}

			address += address3
		}

		fullname := orgtype

		if len(fullname) > 0 {
			fullname += " - "
		}

		fullname += name

		if len(phone) > 0 {
			phone = "phone: " + phone + "\n"
		}

		if len(email) > 0 {
			email = "email: " + email + "\n"
		}

		if len(website) > 0 {
			website = "website: " + website + "\n"
		}

		if len(fax) > 0 {
			fax = "fax: " + fax + "\n"
		}

		description := fmt.Sprintf(
			"%s\n%s %s\n%s%s%s%s",
			fullname,
			address,
			postcode,
			phone,
			email,
			website,
			fax,
		)

		m := &model.Location{
			Type:        geo.PointType_POINT_CLINIC,
			Name:        name,
			Description: description,
			Time:        time.Now(),
			Lon:         lon,
			Lat:         lat,
			Source:      imp.Source(),
			Importer:    imp.ID(),
		}

		return imp.repository.LocationInsert(ctx, m)
	})
}
