package datagovuk

import (
	"context"
	"fmt"
	"net/http"
	"strconv"
	"strings"
	"time"

	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/util"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type importerHospitalsImpl struct {
	repository model.Repository
	client     *http.Client
}

// NewHospitals returns new importer instance
func NewHospitals(repo model.Repository, client *http.Client) importer.Importer {
	return &importerHospitalsImpl{
		repository: repo,
		client:     client,
	}
}

func (imp *importerHospitalsImpl) ID() string {
	return "mashroom-data-gov-uk-hospitals"
}

func (imp *importerHospitalsImpl) Source() string {
	return "https://media.nhswebsite.nhs.uk/data/foi/Hospital.csv"
}

func (imp *importerHospitalsImpl) Import(ctx context.Context) (success, failure int, err error) {
	resp, err := imp.client.Get(imp.Source())
	if err != nil {
		return
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	return util.ImportSplit(resp.Body, "\xAC", func(record []string) (err error) {
		if len(record) < 22 {
			return util.ErrTooShort
		}

		orgtype := record[2]
		name := record[7]
		address1 := record[8]
		address2 := record[9]
		address3 := record[10]
		postcode := record[13]
		latitude := record[14]
		longitude := record[15]
		parentName := record[17]
		phone := record[18]
		email := record[19]
		website := record[20]
		fax := record[21]

		lon, err := strconv.ParseFloat(longitude, 64)
		if err != nil {
			return err
		}

		lat, err := strconv.ParseFloat(latitude, 64)
		if err != nil {
			return err
		}

		var addressparts, fullnameparts []string

		if len(address1) > 0 {
			addressparts = append(addressparts, address1)
		}

		if len(address2) > 0 {
			addressparts = append(addressparts, address2)
		}

		if len(address3) > 0 {
			addressparts = append(addressparts, address3)
		}

		fullnameparts = append(fullnameparts, orgtype)

		if len(parentName) > 0 {
			fullnameparts = append(fullnameparts, parentName)
		}

		fullname := strings.Join(fullnameparts, " ")

		if len(fullname) > 0 {
			fullname += " - "
		}

		fullname += name

		if len(phone) > 0 {
			phone = "phone: " + phone + "\n"
		}

		if len(email) > 0 {
			email = "email: " + email + "\n"
		}

		if len(website) > 0 {
			website = "website: " + website + "\n"
		}

		if len(fax) > 0 {
			fax = "fax: " + fax + "\n"
		}

		description := fmt.Sprintf(
			"%s\n%s %s\n%s%s%s%s",
			fullname,
			strings.Join(addressparts, " "),
			postcode,
			phone,
			email,
			website,
			fax,
		)

		m := &model.Location{
			Type:        geo.PointType_POINT_HOSPITAL,
			Name:        name,
			Description: description,
			Time:        time.Now(),
			Lon:         lon,
			Lat:         lat,
			Source:      imp.Source(),
			Importer:    imp.ID(),
		}

		return imp.repository.LocationInsert(ctx, m)
	})
}
