package datagovuk

import (
	"archive/zip"
	"bytes"
	"context"
	"fmt"
	"io"
	"net/http"
	"strconv"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/util"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

var errNotFound = grpcerr.New(grpcerr.NotFound, "filename not found")

type importerNaptanImpl struct {
	repository model.Repository
	client     *http.Client
}

// NewNaptan returns new importer instance
func NewNaptan(repo model.Repository, client *http.Client) importer.Importer {
	return &importerNaptanImpl{
		repository: repo,
		client:     client,
	}
}

func (imp *importerNaptanImpl) ID() string {
	return "mashroom-data-gov-uk-naptan"
}

func (imp *importerNaptanImpl) Source() string {
	return "http://naptan.app.dft.gov.uk/datarequest/GTFS.ashx"
}

func (imp *importerNaptanImpl) Import(ctx context.Context) (success, failure int, err error) {
	resp, err := imp.client.Get(imp.Source())
	if err != nil {
		return
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	bs, err := io.ReadAll(resp.Body)
	if err != nil {
		return
	}

	zipReader, err := zip.NewReader(bytes.NewReader(bs), int64(len(bs)))
	if err != nil {
		return
	}

	var r io.ReadCloser

	for _, f := range zipReader.File {
		if f.Name == "Stops.txt" {
			r, err = f.Open()

			break
		}
	}

	if r == nil {
		err = errNotFound
		return
	}

	defer func() {
		_ = r.Close()
	}()

	return util.ImportCSV(r, '\t', func(record []string) (err error) {
		if len(record) < 7 {
			return util.ErrTooShort
		}

		stopCode := record[1]
		name := record[2]
		latitude := record[3]
		longitude := record[4]
		vehicleTypeStr := record[6]

		lon, err := strconv.ParseFloat(longitude, 64)
		if err != nil {
			return err
		}

		lat, err := strconv.ParseFloat(latitude, 64)
		if err != nil {
			return err
		}

		description := fmt.Sprintf("stop code: %s, vehicle type: %s", stopCode, vehicleTypeStr)

		var vehicleType geo.PointType

		switch vehicleTypeStr {
		case "1":
			vehicleType = geo.PointType_POINT_METRO_STATION
		case "2":
			vehicleType = geo.PointType_POINT_RAILWAY_STATION
		case "3":
			vehicleType = geo.PointType_POINT_BUS_STATION
		case "4":
			vehicleType = geo.PointType_POINT_FERRY_STATION
		case "21":
			vehicleType = geo.PointType_POINT_PARKING
		case "22":
			vehicleType = geo.PointType_POINT_AIRPORT
		default:
			vehicleType = geo.PointType_POINT_NAPTAN
		}

		m := &model.Location{
			Type:        vehicleType,
			Name:        name,
			Description: description,
			Time:        time.Now(),
			Lon:         lon,
			Lat:         lat,
		}

		return imp.repository.LocationInsert(ctx, m)
	})
}
