//go:generate go-bindata -pkg postcodedistricts -ignore .*\.go .

// Package postcodedistricts provides importer from csv of uk postcodes
package postcodedistricts

import (
	"bytes"
	"context"
	"strconv"
	"strings"

	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/util"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type importerImpl struct {
	repository model.Repository
}

// New returns new importer instance
func New(repo model.Repository) importer.Importer {
	return &importerImpl{
		repository: repo,
	}
}

func (imp *importerImpl) ID() string {
	return "mashroom-postcodes-districts"
}

func (imp *importerImpl) Source() string {
	return "postcodes-districts.csv"
}

func (imp *importerImpl) Import(ctx context.Context) (success, failure int, err error) {
	bs, err := Asset(imp.Source())
	if err != nil {
		return
	}

	return util.ImportCSV(bytes.NewReader(bs), ',', func(record []string) (err error) {
		if len(record) < 13 {
			return util.ErrTooShort
		}

		lat, err := strconv.ParseFloat(record[1], 64)
		if err != nil {
			return err
		}

		lon, err := strconv.ParseFloat(record[2], 64)
		if err != nil {
			return err
		}

		var nearbyPostcodes []*model.PostcodeNearby

		if record[12] != "" {
			nearbyPostcodesArr := strings.Split(record[12], ",")
			if len(nearbyPostcodesArr) > 0 {
				for _, nearbyPostcode := range nearbyPostcodesArr {
					nearbyPostcodes = append(nearbyPostcodes, &model.PostcodeNearby{
						Postcode:       record[0],
						PostcodeNearby: strings.TrimSpace(nearbyPostcode),
					})
				}
			}
		}

		m := &model.Postcode{
			Postcode: record[0],
			Lat:      lat,
			Lon:      lon,
			Name:     record[6],
			County:   record[7],
			Country:  "",
			Region:   "",
		}

		err = imp.repository.PostcodeSave(ctx, m)
		if err != nil {
			return err
		}

		for _, n := range nearbyPostcodes {
			err = imp.repository.PostcodeNearbySave(ctx, &model.PostcodeNearby{
				Postcode:       n.Postcode,
				PostcodeNearby: n.PostcodeNearby,
			})
			if err != nil {
				return err
			}
		}

		return nil
	})
}
