// Package tourpedia provides importers from tour-pedia.org
package tourpedia

import (
	"context"
	"fmt"
	"net/http"
	"strconv"
	"time"

	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/importer/util"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type importerLondonPOIImpl struct {
	repository model.Repository
	client     *http.Client
}

// NewLondonPOI returns new importer instance
func NewLondonPOI(repo model.Repository, client *http.Client) importer.Importer {
	return &importerLondonPOIImpl{
		repository: repo,
		client:     client,
	}
}

func (imp *importerLondonPOIImpl) ID() string {
	return "mashroom-tour-pedia-london-poi"
}

func (imp *importerLondonPOIImpl) Source() string {
	return "http://tour-pedia.org/download/london-poi.csv"
}

func (imp *importerLondonPOIImpl) Import(ctx context.Context) (success, failure int, err error) {
	resp, err := imp.client.Get(imp.Source())
	if err != nil {
		return
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	return util.ImportCSV(resp.Body, ',', func(record []string) (err error) {
		if len(record) < 12 {
			return util.ErrTooShort
		}

		name := record[6]
		latitude := record[3]
		longitude := record[4]
		category := record[9]

		lon, err := strconv.ParseFloat(longitude, 64)
		if err != nil {
			return err
		}

		lat, err := strconv.ParseFloat(latitude, 64)
		if err != nil {
			return err
		}

		m := &model.Location{
			Type:        geo.PointType_POINT_LONDON_POI,
			Name:        name,
			Description: fmt.Sprintf("category: %s", category),
			Lon:         lon,
			Lat:         lat,
			Source:      imp.Source(),
			Importer:    imp.ID(),
			Time:        time.Now(),
		}

		return imp.repository.LocationInsert(ctx, m)
	})
}
