// Package util provides helpers for importers
package util

import (
	"bufio"
	"bytes"
	"encoding/csv"
	"errors"
	"io"
	"net/http"
	"strconv"
	"strings"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
)

// ErrTooShort is returned when dataset in row is not meeting requirements
var ErrTooShort = grpcerr.New(grpcerr.InvalidData, "too short")

// Processor function signature
type Processor func(row []string) (err error)

// ImportSplit processes input splitting rows by newlines and fields by separator character
func ImportSplit(r io.Reader, sep string, proc Processor) (success, failure int, err error) {
	scanner := bufio.NewScanner(r)

	firstLine := true

	for scanner.Scan() {
		if firstLine {
			firstLine = false

			continue
		}

		line := scanner.Text()
		row := strings.Split(line, sep)

		err = proc(row)
		if err != nil {
			log.WithError(err).Warn("Importing split row")

			failure++

			continue
		}

		success++
	}

	return
}

// ImportCSV processes CSV-formatted input
func ImportCSV(r io.Reader, sep rune, proc Processor) (success, failure int, err error) {
	csvReader := csv.NewReader(r)
	csvReader.Comma = sep

	for {
		var row []string

		row, err = csvReader.Read()
		if errors.Is(err, io.EOF) {
			err = nil
			break
		}

		if err != nil {
			return
		}

		err = proc(row)
		if err != nil {
			log.WithError(err).Warn("Importing CSV")

			failure++

			continue
		}

		success++
	}

	return
}

// MockTripper provides http round-tripper mocking in tests
type MockTripper struct {
	ResponseHeaders http.Header
	ResponseCode    int
	ResponseBody    []byte
}

type readerCloser struct {
	io.Reader
}

func (readerCloser) Close() error {
	return nil
}

// RoundTrip returns response based on mocked parameters
func (mock *MockTripper) RoundTrip(req *http.Request) (resp *http.Response, err error) {
	headers := mock.ResponseHeaders
	if headers == nil {
		headers = http.Header{}
	}

	headers.Set("content-length", strconv.FormatInt(int64(len(mock.ResponseBody)), 10))

	return &http.Response{
		Status:        http.StatusText(mock.ResponseCode),
		StatusCode:    mock.ResponseCode,
		Proto:         "HTTP/1.1",
		ProtoMajor:    1,
		ProtoMinor:    1,
		Header:        mock.ResponseHeaders,
		Body:          readerCloser{Reader: bytes.NewReader(mock.ResponseBody)},
		ContentLength: int64(len(mock.ResponseBody)),
		Close:         true,
		Request:       req,
	}, nil
}
