package postgres

import (
	"context"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type repositoryImplCounty struct {
	list struct {
		all             *sqlx.Stmt
		byQuery         *sqlx.Stmt
		byCountryRegion *sqlx.Stmt
		withinRadius    *sqlx.Stmt
	}
}

func prepareCounty(db *sqlx.DB) (stmt repositoryImplCounty, err error) {
	stmt.list.all, err = preparex(db, queryCountyList)
	if err != nil {
		return
	}

	stmt.list.byQuery, err = preparex(db, queryCountyListByQuery)
	if err != nil {
		return
	}

	stmt.list.byCountryRegion, err = preparex(db, queryCountyListByCountryRegion)
	if err != nil {
		return
	}

	stmt.list.withinRadius, err = preparex(db, queryCountyListWithinRadius)
	if err != nil {
		return
	}

	return
}

const queryCountyList = `
select distinct 
  county,
  country,
  region,
  st_x(ST_Centroid(ST_Collect(pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(pos::geometry))) as lat
from towns
group by county,country,region
order by county
offset $1 limit $2
`

func (repo *repositoryImplCounty) CountyList(ctx context.Context, offset, limit int) (ms []*model.County, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.all.SelectContext(ctx, &ms, offset, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryCountyListByQuery = `
select distinct 
  county,
  country,
  region,
  st_x(ST_Centroid(ST_Collect(pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(pos::geometry))) as lat
from towns where 
  initcap(replace(replace(county,'/',' '),'-',' ')) ilike '' || initcap(replace(replace($1 ,'/',' ') ,'-',' ')) || '%'
group by county,country,region
order by county limit $2
`

func (repo *repositoryImplCounty) CountyListByQuery(
	ctx context.Context,
	query string,
	limit int,
) (ms []*model.County, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.byQuery.SelectContext(ctx, &ms, query, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryCountyListByCountryRegion = `
select distinct
  county
from towns where 
  country = $1 and 
  region = $2
limit $3
`

func (repo *repositoryImplCounty) CountyListByCountryRegion(
	ctx context.Context,
	country, region string,
	limit int,
) (ms []*model.County, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.byCountryRegion.SelectContext(ctx, &ms, country, region, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryCountyListWithinRadius = `
with
aabb as (
  select * from towns where
    mx > ($1 * 111111.11) - $3
and my > ($2 * 111111.11) - $3
and mx < ($1 * 111111.11) + $3
and my < ($2 * 111111.11) + $3
),
foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11,
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / $4) * $4 as mx,
    (my / $4) * $4 as my
  from aabb group by 
    (mx / $4) * $4,
    (my / $4) * $4
) select distinct
  coalesce(loc.county, '') as county,
  coalesce(loc.country, '') as country,
  coalesce(loc.region, '') as region,
      st_x(ST_Centroid(ST_Collect(loc.pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(loc.pos::geometry))) as lat
from foo left join aabb loc on
    foo.count < $5
and foo.mx = (loc.mx / $4) * $4
and foo.my = (loc.my / $4) * $4
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3) and
  loc.area != '' and coalesce(loc.area, '') != ''
group by loc.county, loc.region, loc.country
limit $6
`

func (repo *repositoryImplCounty) CountyListWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusLonLat,
	limit int,
) (ms []*model.County, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.withinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
		limit,
	)
	if err != nil {
		return nil, err
	}

	return
}
