package postgres

import (
	"context"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type repositoryImplLocation struct {
	db     *sqlx.DB
	insert *sqlx.NamedStmt
	update *sqlx.NamedStmt
	get    struct {
		byID *sqlx.Stmt
	}
	list struct {
		withinRadius             *sqlx.Stmt
		withinBox                *sqlx.Stmt
		relativeWithinRadius     *sqlx.Stmt
		nearestLocationsAllTypes *sqlx.Stmt
	}
}

func prepareLocation(db *sqlx.DB) (stmt repositoryImplLocation, err error) {
	stmt.db = db

	stmt.insert, err = prepareNamed(db, queryLocationInsert)
	if err != nil {
		return
	}

	stmt.update, err = prepareNamed(db, queryLocationUpdate)
	if err != nil {
		return
	}

	stmt.get.byID, err = preparex(db, queryLocationGetByID)
	if err != nil {
		return
	}

	stmt.list.withinRadius, err = preparex(db, queryLocationListWithinRadius)
	if err != nil {
		return
	}

	stmt.list.withinBox, err = preparex(db, queryLocationListWithinBox)
	if err != nil {
		return
	}

	stmt.list.relativeWithinRadius, err = preparex(db, queryLocationListRelativeWithinRadius)
	if err != nil {
		return
	}

	stmt.list.nearestLocationsAllTypes, err = preparex(db, queryGetNearestLocationsAllTypes)
	if err != nil {
		return
	}

	return
}

const queryLocationInsert = `
insert into locations(
  type,
  name,
  description,
  source,
  importer,
  time,
  pos,
  mx,
  my
) values (
  :type,
  :name,
  :description,
  :source,
  :importer,
  now(),
  ST_SetSRID(ST_MakePoint(:lon, :lat), 4326),
  :lon * 111111.11,
  :lat * 111111.11
) on conflict (name, pos) do UPDATE SET type = :type returning
  id,
  type,
  name,
  description,
  source,
  importer,
  time,
  :lon as lon,
  :lat as lat,
  1 as count
`

func (repo *repositoryImplLocation) LocationInsert(ctx context.Context, location *model.Location) error {
	err := repo.insert.GetContext(ctx, location, location)

	return err
}

const queryLocationUpdate = `
update locations set
  type        = :type,
  name        = :name,
  description = :description,
  source      = :source,
  importer    = :importer,
  time        = now(),
  pos         = ST_SetSRID(ST_MakePoint(:lon, :lat), 4326),
  mx          = :lon * 111111.11,
  my          = :lat * 111111.11 
where id = :id returning
  id,
  type,
  name,
  description,
  source,
  importer,
  time,
  :lon as lon,
  :lat as lat,
  1 as count
`

func (repo *repositoryImplLocation) LocationUpdate(ctx context.Context, location *model.Location) error {
	err := repo.update.GetContext(ctx, location, location)

	return err
}

const queryLocationGetByID = `
select
  1 as count,
  id,
  type, 
  name, 
  description,
  source,
  importer,
  time,
  st_x(pos::geometry) as lon,
  st_y(pos::geometry) as lat
from locations where id = $1 
`

func (repo *repositoryImplLocation) LocationGetByID(ctx context.Context, id uint64) (m *model.Location, err error) {
	err = repo.get.byID.GetContext(ctx, m, id)
	if err != nil {
		return nil, err
	}

	return
}

const queryLocationListWithinRadius = `
with
aabb as (
  select * from locations where
    mx > ($1 * 111111.11) - $3
and my > ($2 * 111111.11) - $3
and mx < ($1 * 111111.11) + $3
and my < ($2 * 111111.11) + $3
),
foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11,
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / $4) * $4 as mx,
    (my / $4) * $4 as my
  from aabb group by 
    (mx / $4) * $4,
    (my / $4) * $4
) select
  case when foo.count < $5 then 1 else foo.count end as count,
  coalesce(loc.id, 0) as id,
  coalesce(loc.type, 0) as type,
  coalesce(loc.name, '') as name,
  coalesce(loc.description, '') as description,
  coalesce(loc.source, '') as source,
  coalesce(loc.importer, '') as importer,
  coalesce(loc.time, now()) as time,
  st_x(coalesce(loc.pos, foo.pos)::geometry) as lon,
  st_y(coalesce(loc.pos, foo.pos)::geometry) as lat
from foo left join aabb loc on
    foo.count < $5
and foo.mx = (loc.mx / $4) * $4
and foo.my = (loc.my / $4) * $4
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
`

func (repo *repositoryImplLocation) LocationListWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusLonLat,
) (ms []*model.Location, err error) {
	err = repo.list.withinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
	)
	if err != nil {
		return nil, err
	}

	return
}

const queryLocationListWithinBox = `
with
aabb as (
  select * from locations where
    mx > $1 * 111111.11
and my > $2 * 111111.11
and mx < $3 * 111111.11
and my < $4 * 111111.11
),
foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11,
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / (($3 - $1) * 111111.11 / $5)::integer) * (($3 - $1) * 111111.11 / $5)::integer as mx,
    (my / (($4 - $2) * 111111.11 / $5)::integer) * (($4 - $2) * 111111.11 / $5)::integer as my
  from aabb group by 
    (mx / (($3 - $1) * 111111.11 / $5)::integer) * (($3 - $1) * 111111.11 / $5)::integer,
    (my / (($4 - $2) * 111111.11 / $5)::integer) * (($4 - $2) * 111111.11 / $5)::integer
) select
  case when foo.count < $6 then 1 else foo.count end as count,
  coalesce(loc.id, 0) as id,
  coalesce(loc.type, 0) as type,
  coalesce(loc.name, '') as name,
  coalesce(loc.description, '') as description,
  coalesce(loc.source, '') as source,
  coalesce(loc.importer, '') as importer,
  coalesce(loc.time, now()) as time,
  st_x(coalesce(loc.pos, foo.pos)::geometry) as lon,
  st_y(coalesce(loc.pos, foo.pos)::geometry) as lat
from foo left join aabb loc on
    foo.count < $6
and foo.mx = (loc.mx / (($3 - $1) * 111111.11 / $5)::integer) * (($3 - $1) * 111111.11 / $5)::integer
and foo.my = (loc.my / (($4 - $2) * 111111.11 / $5)::integer) * (($4 - $2) * 111111.11 / $5)::integer
`

func (repo *repositoryImplLocation) LocationListWithinBox(
	ctx context.Context,
	box *model.AABBLonLat,
) (ms []*model.Location, err error) {
	err = repo.list.withinBox.SelectContext(
		ctx,
		&ms,
		box.AALon,
		box.AALat,
		box.BBLon,
		box.BBLat,
		box.Mod,
		box.Cluster,
	)
	if err != nil {
		return nil, err
	}

	return
}

const queryLocationListRelativeWithinRadius = `
with 
aabb as ( 
  select * from locations where 
    mx > ($1 * 111111.11) - $3 
and my > ($2 * 111111.11) - $3 
and mx < ($1 * 111111.11) + $3 
and my < ($2 * 111111.11) + $3 
and type = $6  
),
foo as ( 
  select 
    count(*), 
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11,
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos, 
    (mx / $4) * $4 as mx, 
    (my / $4) * $4 as my 
  from aabb group by 
    (mx / $4) * $4, 
    (my / $4) * $4 
) select 
  coalesce(loc.id, 0) as id, 
  coalesce(loc.type, 0) as type, 
  coalesce(loc.name, '') as name, 
  coalesce(loc.description, '') as description, 
  st_x(coalesce(loc.pos, foo.pos)::geometry) as lon, 
  st_y(coalesce(loc.pos, foo.pos)::geometry) as lat, 
  ST_Distance(ST_MakePoint($1, $2), loc.pos) as distance 
from foo left join aabb loc on 
    foo.count < $5 
and foo.mx = (loc.mx / $4) * $4 
and foo.my = (loc.my / $4) * $4 
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)  
  and loc.id > 0 
  order by distance limit $7 
`

func (repo *repositoryImplLocation) LocationListRelativeWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusTypeLonLat,
) (ms []*model.LocationRelative, err error) {
	err = repo.list.relativeWithinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
		lonlat.Type,
		lonlat.Limit,
	)
	if err != nil {
		return nil, err
	}

	return
}

const (
	queryGetNearestLocations = `
SELECT id,
       name,
       type,
       description,
       st_x(locations.pos::geometry) as lon,
       st_y(locations.pos::geometry) as lat,
       st_distancesphere(locations.pos::geometry,ST_MakePoint(?, ?)) as distance
FROM locations
WHERE ST_DWithin(locations.pos, ST_SetSRID(ST_MakePoint(?, ?), 4326), ?)
  AND type IN (?)
ORDER BY locations.pos <-> ST_SetSRID(ST_MakePoint(?, ?), 4326)`

	queryGetNearestLocationsAllTypes = `
SELECT id,
       name,
       type,
       description,
       st_x(locations.pos::geometry) as lon,
       st_y(locations.pos::geometry) as lat,
       st_distancesphere(locations.pos::geometry,ST_MakePoint($1, $2)) as distance
FROM locations
WHERE ST_DWithin(locations.pos, ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
ORDER BY locations.pos <-> ST_SetSRID(ST_MakePoint($1, $2), 4326)`
)

func (repo *repositoryImplLocation) GetNearestLocationList(
	ctx context.Context,
	lonLatList []model.LonLatWithLocationTypes,
) (ms [][]*model.LocationRelative, err error) {
	if len(lonLatList) == 0 {
		return
	}

	ms = make([][]*model.LocationRelative, len(lonLatList))

	for i, lonLat := range lonLatList {
		var locations = make([]*model.LocationRelative, 0)

		if len(lonLat.PointTypes) == 0 {
			err = repo.list.nearestLocationsAllTypes.SelectContext(
				ctx,
				&locations,
				lonLat.Lon,
				lonLat.Lat,
				lonLat.Radius,
			)
			if err != nil {
				return nil, err
			}

			ms[i] = locations

			continue
		}

		qq, args, err := sqlx.In(
			queryGetNearestLocations,
			lonLat.Lon,
			lonLat.Lat,
			lonLat.Lon,
			lonLat.Lat,
			lonLat.Radius,
			lonLat.PointTypes,
			lonLat.Lon,
			lonLat.Lat,
		)
		if err != nil {
			return nil, err
		}

		qq = sqlx.Rebind(sqlx.DOLLAR, qq)

		err = repo.db.SelectContext(ctx, &locations, qq, args...)
		if err != nil {
			return nil, err
		}

		ms[i] = locations
	}

	return ms, nil
}
