package postgres

import (
	"context"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type repositoryImplPostcode struct {
	save       *sqlx.NamedStmt
	nearbySave *sqlx.NamedStmt
	list       struct {
		byQuery      *sqlx.Stmt
		withinRadius *sqlx.Stmt
	}
}

func preparePostcode(db *sqlx.DB) (stmt repositoryImplPostcode, err error) {
	stmt.save, err = prepareNamed(db, queryPostcodeSave)
	if err != nil {
		return
	}

	stmt.nearbySave, err = prepareNamed(db, queryPostcodeNearbySave)
	if err != nil {
		return
	}

	stmt.list.byQuery, err = preparex(db, queryPostcodeByQuery)
	if err != nil {
		return
	}

	stmt.list.withinRadius, err = preparex(db, queryPostcodeWithinRadius)
	if err != nil {
		return
	}

	return
}

const queryPostcodeSave = `
insert into postcodes(
  postcode,
  name,
  county,
  region,
  country,
  pos,
  mx,
  my,
  time
) values (
  :postcode,
  :name,
  :county,
  :region,
  :country,
  ST_SetSRID(ST_MakePoint(:lon, :lat), 4326),
  :lon * 111111.11,
  :lat * 111111.11,
  now()
) on CONFLICT (postcode, pos) DO UPDATE SET 
  name = :name,
  county = :county,
  region = :region,  
  country = :country,
  time = now()
returning
  id,
  postcode,
  name,
  county,
  region,
  country,
  :lon as lon,
  :lat as lat
`

func (repo *repositoryImplPostcode) PostcodeSave(ctx context.Context, postcode *model.Postcode) error {
	err := repo.save.GetContext(ctx, postcode, postcode)

	return err
}

const queryPostcodeNearbySave = `
insert into nearbypostcodes(
  postcode,
  nearby,
  time
) values (
  :postcode,
  :nearby,
  now()
) on CONFLICT DO NOTHING
returning
  id,
  postcode,
  nearby
`

func (repo *repositoryImplPostcode) PostcodeNearbySave(ctx context.Context, postcode *model.PostcodeNearby) error {
	err := repo.nearbySave.GetContext(ctx, postcode, postcode)

	return err
}

const queryPostcodeByQuery = `
select distinct 
  postcode,
  name,
  county,
  region,
  country,
  st_x(ST_Centroid(ST_Collect(pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(pos::geometry))) as lat
from postcodes where postcode ilike '' || $1 || '%'
  group by postcode,name,county,region,country
  order by postcode limit $2
`

func (repo *repositoryImplPostcode) PostcodeListByQuery(
	ctx context.Context,
	query string,
	limit int,
) (ms []*model.Postcode, err error) {
	err = repo.list.byQuery.SelectContext(ctx, &ms, query, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryPostcodeWithinRadius = `
with
aabb as (
  select * from postcodes where
    mx > ($1 * 111111.11) - $3
and my > ($2 * 111111.11) - $3
and mx < ($1 * 111111.11) + $3
and my < ($2 * 111111.11) + $3
),
foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11, 
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / $4) * $4 as mx,
    (my / $4) * $4 as my
  from aabb group by 
    (mx / $4) * $4,
    (my / $4) * $4
) select distinct
  coalesce(loc.postcode, '') as postcode,
  coalesce(loc.name, '') as name,
  coalesce(loc.county, '') as county,
  coalesce(loc.region, '') as region,
  coalesce(loc.country, '') as country,
  st_x(ST_Centroid(ST_Collect(loc.pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(loc.pos::geometry))) as lat
from foo left join aabb loc on
    foo.count < $5
and foo.mx = (loc.mx / $4) * $4
and foo.my = (loc.my / $4) * $4
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
  and loc.postcode != '' and coalesce(loc.postcode, '') != ''
  group by loc.postcode, loc.name, loc.county, loc.region, loc.country
  limit $6
`

func (repo *repositoryImplPostcode) PostcodeListWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusLonLat,
	limit int,
) (ms []*model.Postcode, err error) {
	err = repo.list.withinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
		limit,
	)
	if err != nil {
		return nil, err
	}

	return
}
