package postgres

import (
	"context"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type repositoryImplRegion struct {
	list struct {
		all          *sqlx.Stmt
		byQuery      *sqlx.Stmt
		byCountry    *sqlx.Stmt
		withinRadius *sqlx.Stmt
	}
}

func prepareRegion(db *sqlx.DB) (stmt repositoryImplRegion, err error) {
	stmt.list.all, err = preparex(db, queryRegionList)
	if err != nil {
		return
	}

	stmt.list.byQuery, err = preparex(db, queryRegionListByQuery)
	if err != nil {
		return
	}

	stmt.list.byCountry, err = preparex(db, queryRegionListByCountry)
	if err != nil {
		return
	}

	stmt.list.withinRadius, err = preparex(db, queryRegionListWithinRadius)
	if err != nil {
		return
	}

	return
}

const queryRegionList = `
select distinct 
  country,
  region,
  st_x(ST_Centroid(ST_Collect(pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(pos::geometry))) as lat
from towns
group by region, country
order by region
offset $1 limit $2
`

func (repo *repositoryImplRegion) RegionList(ctx context.Context, offset, limit int) (ms []*model.Region, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.all.SelectContext(ctx, &ms, offset, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryRegionListByQuery = `
select distinct 
  country,
  region,
  st_x(ST_Centroid(ST_Collect(pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(pos::geometry))) as lat
from towns where
  initcap(replace(replace(region,'/',' '),'-',' ')) ilike '' || initcap(replace(replace($1 ,'/',' ') ,'-',' ')) || '%'
group by region,country
order by region limit $2
`

func (repo *repositoryImplRegion) RegionListByQuery(
	ctx context.Context,
	query string,
	limit int,
) (ms []*model.Region, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.byQuery.SelectContext(ctx, &ms, query, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryRegionListByCountry = `
select distinct
  region,
  country
from towns where
  country = $1
`

func (repo *repositoryImplRegion) RegionListByCountry(
	ctx context.Context,
	country string,
) (ms []*model.Region, err error) {
	err = repo.list.byCountry.SelectContext(ctx, &ms, country)
	if err != nil {
		return nil, err
	}

	return
}

const queryRegionListWithinRadius = `
with
aabb as (
  select * from towns where
    mx > ($1 * 111111.11) - $3
and my > ($2 * 111111.11) - $3
and mx < ($1 * 111111.11) + $3
and my < ($2 * 111111.11) + $3
),
foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11, 
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / $4) * $4 as mx,
    (my / $4) * $4 as my
  from aabb group by 
    (mx / $4) * $4,
    (my / $4) * $4
) select distinct
  coalesce(loc.country, '') as country,
  coalesce(loc.region, '') as region,
      st_x(ST_Centroid(ST_Collect(loc.pos::geometry))) as lon,
  st_y(ST_Centroid(ST_Collect(loc.pos::geometry))) as lat
from foo left join aabb loc on
    foo.count < $5
and foo.mx = (loc.mx / $4) * $4
and foo.my = (loc.my / $4) * $4
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
  and loc.region != '' and coalesce(loc.region, '') != ''
  group by loc.region, loc.country
  limit $6
`

func (repo *repositoryImplRegion) RegionListWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusLonLat,
	limit int,
) (ms []*model.Region, err error) {
	err = repo.list.withinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
		limit,
	)
	if err != nil {
		return nil, err
	}

	return
}
