package postgres

import (
	"context"
	"database/sql"

	"github.com/jmoiron/sqlx"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
)

type repositoryImplTown struct {
	insert *sqlx.NamedStmt
	update *sqlx.NamedStmt
	get    struct {
		byID        *sqlx.Stmt
		nearestTown *sqlx.Stmt
	}
	list struct {
		all                   *sqlx.Stmt
		sub                   *sqlx.Stmt
		byQuery               *sqlx.Stmt
		byCountry             *sqlx.Stmt
		byCountryRegionCounty *sqlx.Stmt
		withinRadius          *sqlx.Stmt
	}
}

func prepareTown(db *sqlx.DB) (stmt repositoryImplTown, err error) {
	stmt.insert, err = prepareNamed(db, queryTownInsert)
	if err != nil {
		return
	}

	stmt.update, err = prepareNamed(db, queryTownUpdate)
	if err != nil {
		return
	}

	stmt.get.byID, err = preparex(db, queryTownGetByID)
	if err != nil {
		return
	}

	stmt.get.nearestTown, err = preparex(db, queryGetNearestTown)
	if err != nil {
		return
	}

	stmt.list.all, err = preparex(db, queryTownList)
	if err != nil {
		return
	}

	stmt.list.sub, err = preparex(db, queryTownListSub)
	if err != nil {
		return
	}

	stmt.list.byQuery, err = preparex(db, queryTownListByQuery)
	if err != nil {
		return
	}

	stmt.list.byCountry, err = preparex(db, queryTownListByCountry)
	if err != nil {
		return
	}

	stmt.list.byCountryRegionCounty, err = preparex(db, queryTownListByCountryRegionCounty)
	if err != nil {
		return
	}

	stmt.list.withinRadius, err = preparex(db, queryTownListWithinRadius)
	if err != nil {
		return
	}

	return
}

const queryTownInsert = `
insert into towns(
  name,
  county,
  country,
  area,
  region,
  postcode,
  type,
  pos,
  mx,
  my,
  time
) values (
  :name,
  :county,
  :country,
  :area, 
  :region,  
  :postcode,
  :type,
  ST_SetSRID(ST_MakePoint(:lon, :lat), 4326),
  :lon * 111111.11,
  :lat * 111111.11,
  now()
) on conflict (name, pos) do UPDATE SET 
  county = :county, 
  country = :country, 
  area = :area, 
  region = :region, 
  postcode = :postcode,
  type = :type, 
  time = now()
returning
  id,
  name,
  country
  region,
  county,
  area,
  postcode,
  type,
  :lon as lon,
  :lat as lat,
  time
`

func (repo *repositoryImplTown) TownInsert(ctx context.Context, town *model.Town) error {
	err := repo.insert.GetContext(ctx, town, town)

	return err
}

const queryTownUpdate = `
update towns set
  type        = :type,
  name        = :name,
  time        = now(),
  pos         = ST_SetSRID(ST_MakePoint(:lon, :lat), 4326),
  mx          = :lon * 111111.11,
  my          = :lat * 111111.11 
where id = :id
returning
  id,
  name,
  country
  region,
  county,
  area,
  postcode,
  type,
  :lon as lon,
  :lat as lat,
  time
`

func (repo *repositoryImplTown) TownUpdate(ctx context.Context, town *model.Town) error {
	err := repo.update.GetContext(ctx, town, town)

	return err
}

const queryTownGetByID = `
select
  id,
  name,
  county,
  country,
  region,
  postcode,
  type,
  ST_X(pos::geometry) as lon,
  ST_Y(pos::geometry) as lat,
  time
from towns where
  id = $1
`

func (repo *repositoryImplTown) TownGetByID(ctx context.Context, id uint64) (m *model.Town, err error) {
	err = repo.get.byID.GetContext(ctx, m, id)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownList = `
select
  row_number() over(order by name) as ord,
  id,
  name,
  county,
  country,
  region,
  postcode,
  type,
  ST_X(pos::geometry) as lon,
  ST_Y(pos::geometry) as lat,
  time
from towns
order by name
offset $1 limit $2
`

func (repo *repositoryImplTown) TownList(ctx context.Context, offset, limit int) (ms []*model.Town, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.all.SelectContext(ctx, &ms, offset, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownListSub = `
select
  t1.id,
  t1.name,
  t1.county,
  t1.country,
  t1.region,
  t1.postcode,
  t1.type,
  ST_X(t1.pos::geometry) as lon,
  ST_Y(t1.pos::geometry) as lat,
  t1.time
from towns t1 join towns t2 on t1.area = t2.area
where
  t2.id = $1 and 
  t2.type = 'City' and 
  (t1.type = 'City' or t1.type = 'Town' or t1.type='Village' or t1.type='Suburban Area') and 
  t1.id <> $1
order by t1.name limit $2
`

func (repo *repositoryImplTown) TownListSub(ctx context.Context, id uint64, limit int) (ms []*model.Town, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.sub.SelectContext(ctx, &ms, id, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownListByQuery = `
with first as (
  select distinct
    1 as ord,
    id,
    name,
    county,
    country,
    region,
    postcode,
    type,
    ST_X(pos::geometry) as lon,
    ST_Y(pos::geometry) as lat,
    time 
  from towns where
    initcap(replace(replace(name,'/',' '),'-',' ')) ilike '' || initcap(replace(replace($1 ,'/',' ') ,'-',' ')) || '%' 
), second as (
  select distinct
    2 as ord, 
    id,
    name,
    county,
    country,
    region,
    postcode,
    type,
    ST_X(pos::geometry) as lon,
    ST_Y(pos::geometry) as lat,
    time 
  from towns where 
    initcap(replace(replace(name,'/',' '),'-',' ')) ilike '%' || initcap(replace(replace($1 ,'/',' ') ,'-',' ')) || '%' 
) select
  ord,
  id,
  name,
  county,
  country,
  region,
  postcode,
  type,
  lon,
  lat,
  time
from first union
select s.ord,
  s.id,
  s.name,
  s.county,
  s.country,
  s.region,
  s.postcode,
  s.type,
  s.lon,
  s.lat,
  s.time
from second s left join first f on f.name = s.name
where f.name is null
order by ord, name
limit $2
`

func (repo *repositoryImplTown) TownListByQuery(
	ctx context.Context,
	query string,
	limit int,
) (ms []*model.Town, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.byQuery.SelectContext(ctx, &ms, query, limit)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownListByCountry = `
select 
  name,
  county,
  region,
  country
from towns where
  country = $1
order by name
`

func (repo *repositoryImplTown) TownListByCountry(
	ctx context.Context,
	country string,
) (ms []*model.Town, err error) {
	err = repo.list.byCountry.SelectContext(ctx, &ms, country)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownListByCountryRegionCounty = `
select distinct
  "name"
from towns where
  country = $1 and
  region = $2 and
  county = $3
`

func (repo *repositoryImplTown) TownListByCountryRegionCounty(
	ctx context.Context,
	country, region, county string,
) (ms []*model.Town, err error) {
	err = repo.list.byCountryRegionCounty.SelectContext(ctx, &ms, country, region, county)
	if err != nil {
		return nil, err
	}

	return
}

const queryTownListWithinRadius = `
with
aabb as (
  select * from towns where
    mx > ($1 * 111111.11) - $3 and
    my > ($2 * 111111.11) - $3 and
    mx < ($1 * 111111.11) + $3 and
    my < ($2 * 111111.11) + $3
), foo as (
  select
    count(*),
    ST_MakePoint(
      (min(mx) + (max(mx) - min(mx)) / 2) / 111111.11,
      (min(my) + (max(my) - min(my)) / 2) / 111111.11
    ) as pos,
    (mx / $4) * $4 as mx,
    (my / $4) * $4 as my
  from aabb group by 
    (mx / $4) * $4,
    (my / $4) * $4
) select
  coalesce(loc.id, 0) as id,
  coalesce(loc.name, '') as name,
  coalesce(loc.county, '') as county,
  coalesce(loc.country, '') as country,
  coalesce(loc.region, '') as region,
  coalesce(loc.postcode, '') as postcode,
  coalesce(loc.type, '') as type,
  st_x(coalesce(loc.pos, foo.pos)::geometry) as lon,
  st_y(coalesce(loc.pos, foo.pos)::geometry) as lat,
  coalesce(loc.time, now()) as time
from foo left join aabb loc on
  foo.count < $5 and
  foo.mx = (loc.mx / $4) * $4 and
  foo.my = (loc.my / $4) * $4
where
  ST_DWithin(coalesce(loc.pos, foo.pos), ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
  order by name limit $6
`

func (repo *repositoryImplTown) TownListWithinRadius(
	ctx context.Context,
	lonlat *model.RadiusLonLat,
	limit int,
) (ms []*model.Town, err error) {
	if limit == 0 {
		limit = 1 << 30
	}

	err = repo.list.withinRadius.SelectContext(
		ctx,
		&ms,
		lonlat.Lon,
		lonlat.Lat,
		lonlat.Radius/2,
		lonlat.Radius/lonlat.Mod,
		lonlat.Cluster,
		limit,
	)
	if err != nil {
		return nil, err
	}

	return
}

const queryGetNearestTown = `
SELECT 
    id,
    name,
    county,
    country,
    region,
    postcode,
    type,
    st_x(towns.pos::geometry) as lon,
    st_y(towns.pos::geometry) as lat,
	time
FROM towns
WHERE ST_DWithin(towns.pos, ST_SetSRID(ST_MakePoint($1, $2), 4326), $3)
ORDER BY towns.pos <-> ST_SetSRID(ST_MakePoint($1, $2), 4326)
LIMIT 1
`

func (repo *repositoryImplTown) GetNearestTownsForLocationList(
	ctx context.Context,
	radiusLonLatList []model.RadiusLonLat,
) (towns []*model.Town, err error) {
	if len(radiusLonLatList) == 0 {
		return
	}

	towns = make([]*model.Town, len(radiusLonLatList))

	for i, lonLat := range radiusLonLatList {
		var town model.Town

		err = repo.get.nearestTown.GetContext(ctx, &town, lonLat.Lon, lonLat.Lat, lonLat.Radius)
		if err == sql.ErrNoRows {
			continue
		}

		if err != nil {
			return nil, err
		}

		towns[i] = &town
	}

	return towns, nil
}
