package service

import (
	"strconv"

	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
	"gitlab.mashplace.com/mashplace/backend/geo/pkg/mapbox"
)

func (srv *serverImpl) renderPOIs(ms []*mapbox.POI) (pois []*geo.Poi) {
	for _, p := range ms {
		pois = append(pois, srv.renderPOI(p))
	}

	return
}

func (srv *serverImpl) renderPOI(m *mapbox.POI) (poi *geo.Poi) {
	return &geo.Poi{
		Tag:      m.Type,
		Name:     m.Name,
		MakiType: m.MakiType,
		Location: &geo.Location{
			Longitude: m.Location.Longitude,
			Latitude:  m.Location.Latitude,
		},
	}
}

func (srv *serverImpl) renderIsochrones(ms []*mapbox.Isochrone) (isos []*geo.Isochrone) {
	for _, p := range ms {
		isos = append(isos, srv.renderIsochrone(p))
	}

	return
}

func (srv *serverImpl) renderIsochrone(m *mapbox.Isochrone) (iso *geo.Isochrone) {
	return &geo.Isochrone{
		Time:     m.Time,
		Contours: srv.renderContours(m.Contours),
	}
}

func (srv *serverImpl) renderContours(ms []mapbox.Contour) (contours []*geo.Contour) {
	for _, p := range ms {
		contours = append(contours, srv.renderContour(p))
	}

	return
}

func (srv *serverImpl) renderContour(m mapbox.Contour) (contour *geo.Contour) {
	return &geo.Contour{
		Locations: srv.renderLocations(m),
	}
}

func (srv *serverImpl) renderLocations(ms []*mapbox.Location) (locs []*geo.Location) {
	for _, p := range ms {
		locs = append(locs, srv.renderLocation(p))
	}

	return
}

func (srv *serverImpl) renderLocation(m *mapbox.Location) (loc *geo.Location) {
	return &geo.Location{
		Longitude: m.Longitude,
		Latitude:  m.Latitude,
	}
}

func (srv *serverImpl) renderPointsRelative(ms []*model.LocationRelative) (pts []*geo.PointRelative) {
	for _, p := range ms {
		pts = append(pts, srv.renderPointRelative(p))
	}

	return
}

func (srv *serverImpl) renderPointRelative(m *model.LocationRelative) (pt *geo.PointRelative) {
	var id string

	var typ geo.PointType

	if m.ID != 0 {
		id = strconv.FormatUint(m.ID, 10)
		typ = m.Type
	} else {
		typ = geo.PointType_POINT_CLUSTER
	}

	return &geo.PointRelative{
		Point: &geo.Point{
			Id:          id,
			Type:        typ,
			Name:        m.Name,
			Description: m.Description,
			Location: &geo.Location{
				Longitude: m.Lon,
				Latitude:  m.Lat,
			},
			Count: 0,
		},
		Distance: m.Distance,
	}
}

func (srv *serverImpl) renderPoints(ms []*model.Location) (pts []*geo.Point) {
	for _, p := range ms {
		pts = append(pts, srv.renderPoint(p))
	}

	return
}

func (srv *serverImpl) renderPoint(m *model.Location) (pt *geo.Point) {
	var id string

	var typ geo.PointType

	if m.ID != 0 {
		id = strconv.FormatUint(m.ID, 10)
		typ = m.Type
	} else {
		typ = geo.PointType_POINT_CLUSTER
	}

	return &geo.Point{
		Id:          id,
		Type:        typ,
		Name:        m.Name,
		Description: m.Description,
		Location: &geo.Location{
			Longitude: m.Lon,
			Latitude:  m.Lat,
		},
		Count: m.Count,
	}
}

func (srv *serverImpl) renderTowns(ms []*model.Town) (towns []*geo.Town) {
	for _, p := range ms {
		towns = append(towns, srv.renderTown(p))
	}

	return
}

func (srv *serverImpl) renderTown(m *model.Town) (town *geo.Town) {
	if m == nil {
		return nil
	}

	return &geo.Town{
		Id:      strconv.FormatUint(m.ID, 10),
		Name:    m.Name,
		County:  m.County,
		Country: m.Country,
		Region:  m.Region,
		Location: &geo.Location{
			Longitude: m.Lon,
			Latitude:  m.Lat,
		},
		Postcode: m.Postcode,
	}
}

func (srv *serverImpl) renderCounties(ms []*model.County) (counties []*geo.County) {
	for _, p := range ms {
		counties = append(counties, srv.renderCounty(p))
	}

	return
}

func (srv *serverImpl) renderCounty(m *model.County) (county *geo.County) {
	return &geo.County{
		Name:    m.Name,
		Region:  m.Region,
		Country: m.Country,
		Location: &geo.Location{
			Longitude: m.Lon,
			Latitude:  m.Lat,
		},
	}
}

func (srv *serverImpl) renderRegions(ms []*model.Region) (regions []*geo.Region) {
	for _, p := range ms {
		regions = append(regions, srv.renderRegion(p))
	}

	return
}

func (srv *serverImpl) renderRegion(m *model.Region) (region *geo.Region) {
	return &geo.Region{
		Name:    m.Name,
		Country: m.Country,
		Location: &geo.Location{
			Longitude: m.Lon,
			Latitude:  m.Lat,
		},
	}
}

func (srv *serverImpl) renderCountries(ms []*model.Country) (countries []*geo.Country) {
	for _, p := range ms {
		countries = append(countries, srv.renderCountry(p))
	}

	return
}

func (srv *serverImpl) renderCountry(m *model.Country) (region *geo.Country) {
	return &geo.Country{
		Name: m.Name,
	}
}

func (srv *serverImpl) renderPostcodes(ms []*model.Postcode) (postcodes []*geo.Postcode) {
	for _, p := range ms {
		postcodes = append(postcodes, srv.renderPostcode(p))
	}

	return
}

func (srv *serverImpl) renderPostcode(m *model.Postcode) (postcode *geo.Postcode) {
	return &geo.Postcode{
		Code:    m.Postcode,
		Name:    m.Name,
		County:  m.County,
		Region:  m.Region,
		Country: m.Country,
		Location: &geo.Location{
			Longitude: m.Lon,
			Latitude:  m.Lat,
		},
	}
}
