package service

import (
	"context"
	"errors"
	"io"
	"strconv"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"

	"gitlab.mashplace.com/mashplace/backend/common/generic/log"
	geo "gitlab.mashplace.com/mashplace/backend/geo/api"
	"gitlab.mashplace.com/mashplace/backend/geo/internal/model"
	"gitlab.mashplace.com/mashplace/backend/geo/pkg/mapbox"
	"gopkg.in/ugjka/go-tz.v2/tz"
)

var errTooBroad = grpcerr.New(grpcerr.InvalidData, "too broad")
var errBadRequest = grpcerr.New(grpcerr.InvalidData, "bad request")
var errNoCenter = grpcerr.New(grpcerr.InvalidData, "bad request, no center location")
var errNoLocations = grpcerr.New(grpcerr.InvalidData, "bad request, no locations")

func (srv *serverImpl) GetPois(ctx context.Context, req *geo.PoiRequest) (*geo.PoiResponse, error) {
	pois, err := srv.MapboxClient.GetPOI(
		ctx,
		&mapbox.Location{
			Longitude: req.GetLocation().GetLongitude(),
			Latitude:  req.GetLocation().GetLatitude(),
		},
		req.GetTags(),
		req.GetRadiusInMeters(),
	)
	if err != nil {
		log.WithContext(ctx).WithError(err).Warn("Requesting POIs")
	}

	return &geo.PoiResponse{
		Pois: srv.renderPOIs(pois),
	}, nil
}

func (srv *serverImpl) GetIsochrones(ctx context.Context, req *geo.IsochroneRequest) (*geo.IsochroneResponse, error) {
	var periods []time.Duration

	for _, v := range req.GetTimes() {
		periods = append(periods, time.Duration(v)*time.Minute)
	}

	isos, err := srv.MapboxClient.GetIsochrone(
		ctx,
		&mapbox.Location{
			Longitude: req.GetLocation().GetLongitude(),
			Latitude:  req.GetLocation().GetLatitude(),
		},
		periods,
		req.GetMovingType().String(),
		req.GetDenoise(),
	)
	if err != nil {
		log.WithContext(ctx).WithError(err).Warn("Requesting isochrones")
	}

	return &geo.IsochroneResponse{
		Isochrones: srv.renderIsochrones(isos),
	}, nil
}

func (srv *serverImpl) GetPoints(
	ctx context.Context,
	req *geo.GetPointsRequest,
) (resp *geo.GetPointsResponse, err error) {
	var locs []*model.Location

	switch m := req.GetMethod().(type) {
	case *geo.GetPointsRequest_Radius:
		if m.Radius.Meters > 100000 {
			return nil, errTooBroad
		}

		locs, err = srv.Repository.LocationListWithinRadius(ctx, &model.RadiusLonLat{
			Mod:     model.FindWithinMod,
			Cluster: model.FindWithinCluster,
			Lon:     m.Radius.Center.Longitude,
			Lat:     m.Radius.Center.Latitude,
			Radius:  m.Radius.Meters,
		})
	case *geo.GetPointsRequest_Box:
		if m.Box.Bottomright.Longitude-m.Box.Topleft.Longitude > 100000 ||
			m.Box.Bottomright.Latitude-m.Box.Topleft.Latitude > 100000 {
			return nil, errTooBroad
		}

		locs, err = srv.Repository.LocationListWithinBox(ctx, &model.AABBLonLat{
			Mod:     model.FindWithinMod,
			Cluster: model.FindWithinCluster,
			AALon:   m.Box.Topleft.Longitude,
			AALat:   m.Box.Topleft.Latitude,
			BBLon:   m.Box.Bottomright.Longitude,
			BBLat:   m.Box.Bottomright.Latitude,
		})
	}

	if err != nil {
		return nil, err
	}

	return &geo.GetPointsResponse{
		Points: srv.renderPoints(locs),
	}, nil
}

func (srv *serverImpl) GetNearestPoints(
	ctx context.Context,
	req *geo.GetNearestPointsRequest,
) (*geo.GetNearestPointsResponse, error) {
	var lonLatList []model.LonLatWithLocationTypes
	for _, location := range req.GetLocations() {
		lonLatList = append(lonLatList, model.LonLatWithLocationTypes{
			Lon:        location.GetLongitude(),
			Lat:        location.GetLatitude(),
			Radius:     req.GetRadiusInMeters(),
			PointTypes: req.GetPointTypes(),
		})
	}

	result, err := srv.Repository.GetNearestLocationList(ctx, lonLatList)
	if err != nil {
		return nil, err
	}

	resp := &geo.GetNearestPointsResponse{
		PointsForLocation: make([]*geo.PointsForLocation, len(result)),
	}

	for i, points := range result {
		resp.PointsForLocation[i] = &geo.PointsForLocation{
			Points: srv.renderPointsRelative(points),
		}
	}

	return resp, nil
}

func (srv *serverImpl) importUpdate(ctx context.Context) error {
	for _, imp := range srv.Importers {
		log.Infof("Running importer %s on %s", imp.ID(), imp.Source())

		last, _ := srv.Repository.ImportGetLastByImporterSource(ctx, imp.ID(), imp.Source())
		if last != nil && time.Since(last.Time) < srv.Config.MinImportPeriod {
			log.Infof("Reattempt too early %v < %v", time.Since(last.Time), srv.Config.MinImportPeriod)
			continue
		}

		var errstr string

		success, failure, err := imp.Import(ctx)
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("error importing")
			errstr = err.Error()
		}

		log.Infof("results: success:%d failure:%d", success, failure)

		saveerr := srv.Repository.ImportSave(ctx, &model.Import{
			Source:   imp.Source(),
			Importer: imp.ID(),
			Time:     time.Now(),
			Success:  uint64(success),
			Failure:  uint64(failure),
			Message:  errstr,
		})

		if err != nil {
			return err
		}

		if saveerr != nil {
			return saveerr
		}
	}

	return nil
}

func (srv *serverImpl) UpdateDB(ctx context.Context, req *geo.UpdateDBRequest) (*geo.UpdateDBResponse, error) {
	go func() {
		srv.mutex.Lock()
		defer srv.mutex.Unlock()

		err := srv.importUpdate(context.Background())
		if err != nil {
			log.WithContext(ctx).WithError(err).Error("Import updating DB")
		}

		log.Infof("done updating db")
	}()

	return &geo.UpdateDBResponse{}, nil
}

func (srv *serverImpl) GetStations(ctx context.Context, req *geo.GetStationsRequest) (*geo.GetStationsResponse, error) {
	if req.GetRadiusInMeters() > 10000 {
		return nil, errTooBroad
	}

	if req.GetCenter() == nil {
		return nil, errNoCenter
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	points, err := srv.Repository.LocationListRelativeWithinRadius(ctx, &model.RadiusTypeLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster * 10,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
		Type:    req.GetPointType(),
		Limit:   limit,
	})
	if err != nil {
		return nil, err
	}

	return &geo.GetStationsResponse{
		Points: srv.renderPointsRelative(points),
	}, nil
}

func (srv *serverImpl) GetTowns(ctx context.Context, req *geo.GetTownsRequest) (*geo.GetTownsResponse, error) {
	if len(req.GetQuery()) == 0 {
		return nil, errBadRequest
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	towns, err := srv.Repository.TownListByQuery(ctx, req.GetQuery(), int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetTownsResponse{
		Towns: srv.renderTowns(towns),
	}, nil
}

func (srv *serverImpl) GetTownsInRadius(
	ctx context.Context,
	req *geo.GetLocalityInRadiusRequest,
) (*geo.GetTownsResponse, error) {
	if req.GetRadiusInMeters() > 10000 {
		return nil, errTooBroad
	}

	if req.GetCenter() == nil {
		return nil, errNoCenter
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	towns, err := srv.Repository.TownListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetTownsResponse{
		Towns: srv.renderTowns(towns),
	}, nil
}

func (srv *serverImpl) GetNearestTownsForLocationList(
	ctx context.Context,
	req *geo.GetNearestTownsForLocationListRequest,
) (*geo.GetNearestTownsForLocationListResponse, error) {
	if req.GetRadiusInMeters() > 10000 {
		return nil, errTooBroad
	}

	if len(req.GetLocations()) == 0 {
		return nil, errNoLocations
	}

	var locations = make([]model.RadiusLonLat, 0, len(req.GetLocations()))
	for _, loc := range req.GetLocations() {
		locations = append(locations, model.RadiusLonLat{
			Lon:    loc.Longitude,
			Lat:    loc.Latitude,
			Radius: req.RadiusInMeters,
		})
	}

	towns, err := srv.Repository.GetNearestTownsForLocationList(ctx, locations)
	if err != nil {
		return nil, err
	}

	return &geo.GetNearestTownsForLocationListResponse{
		Towns: srv.renderTowns(towns),
	}, nil
}

func (srv *serverImpl) GetCountiesInRadius(
	ctx context.Context,
	req *geo.GetLocalityInRadiusRequest,
) (*geo.GetCountiesResponse, error) {
	if req.GetRadiusInMeters() > 10000 {
		return nil, errTooBroad
	}

	if req.GetCenter() == nil {
		return nil, errNoCenter
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	counties, err := srv.Repository.CountyListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetCountiesResponse{
		Counties: srv.renderCounties(counties),
	}, nil
}

func (srv *serverImpl) GetRegionsInRadius(
	ctx context.Context,
	req *geo.GetLocalityInRadiusRequest,
) (*geo.GetRegionsResponse, error) {
	if req.GetRadiusInMeters() > 10000 {
		return nil, errTooBroad
	}

	if req.GetCenter() == nil {
		return nil, errNoCenter
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	regions, err := srv.Repository.RegionListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetRegionsResponse{
		Regions: srv.renderRegions(regions),
	}, nil
}

func (srv *serverImpl) GetPlacesInRadius(
	ctx context.Context,
	req *geo.GetLocalityInRadiusRequest,
) (*geo.GetLocalityResponse, error) {
	if req.GetRadiusInMeters() > 100000 {
		return nil, errTooBroad
	}

	if req.GetCenter() == nil {
		return nil, errNoCenter
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	towns, err := srv.Repository.TownListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	regions, err := srv.Repository.RegionListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	counties, err := srv.Repository.CountyListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	postcodes, err := srv.Repository.PostcodeListWithinRadius(ctx, &model.RadiusLonLat{
		Mod:     model.FindWithinMod,
		Cluster: model.FindWithinCluster,
		Lon:     req.GetCenter().GetLongitude(),
		Lat:     req.GetCenter().GetLatitude(),
		Radius:  req.GetRadiusInMeters(),
	}, int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetLocalityResponse{
		Towns:     srv.renderTowns(towns),
		Regions:   srv.renderRegions(regions),
		Counties:  srv.renderCounties(counties),
		Postcodes: srv.renderPostcodes(postcodes),
	}, nil
}

func (srv *serverImpl) getPlacesByQueryAll(
	ctx context.Context,
	query string,
) (resp *geo.GetLocalityResponse, err error) {
	parts := strings.Split(query, "/")

	var typ string

	var offset, limit int64

	if len(parts) == 0 {
		return nil, errBadRequest
	}

	typ = strings.ToLower(parts[0])

	if len(parts) > 1 {
		offset, err = strconv.ParseInt(parts[1], 10, 64)
		if err != nil {
			return
		}
	}

	if len(parts) > 2 {
		limit, err = strconv.ParseInt(parts[2], 10, 64)
		if err != nil {
			return
		}
	}

	switch typ {
	case "towns":
		var towns []*model.Town

		towns, err = srv.Repository.TownList(ctx, int(offset), int(limit))
		if err != nil {
			return
		}

		resp = &geo.GetLocalityResponse{
			Towns: srv.renderTowns(towns),
		}
	case "regions":
		var regions []*model.Region

		regions, err = srv.Repository.RegionList(ctx, int(offset), int(limit))
		if err != nil {
			return
		}

		resp = &geo.GetLocalityResponse{
			Regions: srv.renderRegions(regions),
		}
	case "counties":
		var counties []*model.County

		counties, err = srv.Repository.CountyList(ctx, int(offset), int(limit))
		if err != nil {
			return
		}

		resp = &geo.GetLocalityResponse{
			Counties: srv.renderCounties(counties),
		}
	case "countries":
		var countries []*model.Country

		countries, err = srv.Repository.CountryList(ctx, int(offset), int(limit))
		if err != nil {
			return
		}

		resp = &geo.GetLocalityResponse{
			Countries: srv.renderCountries(countries),
		}
	default:
		return &geo.GetLocalityResponse{}, nil
	}

	return
}

func (srv *serverImpl) GetPlacesByQuery(
	ctx context.Context,
	req *geo.GetLocalityRequest,
) (*geo.GetLocalityResponse, error) {
	if len(req.GetQuery()) == 0 {
		return nil, errBadRequest
	}

	if strings.HasPrefix(req.GetQuery(), "/all") {
		return srv.getPlacesByQueryAll(ctx, strings.TrimPrefix(req.GetQuery(), "/all"))
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	towns, err := srv.Repository.TownListByQuery(ctx, req.GetQuery(), int(limit))
	if err != nil {
		return nil, err
	}

	regions, err := srv.Repository.RegionListByQuery(ctx, req.GetQuery(), int(limit))
	if err != nil {
		return nil, err
	}

	counties, err := srv.Repository.CountyListByQuery(ctx, req.GetQuery(), int(limit))
	if err != nil {
		return nil, err
	}

	postcodes, err := srv.Repository.PostcodeListByQuery(ctx, req.GetQuery(), int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetLocalityResponse{
		Towns:     srv.renderTowns(towns),
		Regions:   srv.renderRegions(regions),
		Counties:  srv.renderCounties(counties),
		Postcodes: srv.renderPostcodes(postcodes),
	}, nil
}

func (srv *serverImpl) GetSubTowns(ctx context.Context, req *geo.GetSubTownsRequest) (*geo.GetTownsResponse, error) {
	id, err := strconv.ParseUint(req.GetId(), 10, 64)
	if err != nil {
		return nil, err
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	towns, err := srv.Repository.TownListSub(ctx, id, int(limit))
	if err != nil {
		return nil, err
	}

	return &geo.GetTownsResponse{
		Towns: srv.renderTowns(towns),
	}, nil
}

func (srv *serverImpl) GetCountiesInRegion(
	ctx context.Context,
	req *geo.GetCountiesInRegionRequest,
) (*geo.GetCountiesInRegionResponse, error) {
	if len(req.GetRegion().GetName()) == 0 {
		return nil, errBadRequest
	}

	limit := req.GetLimit()
	if limit <= 0 {
		limit = 3
	}

	counties, err := srv.Repository.CountyListByCountryRegion(
		ctx,
		req.GetRegion().GetCountry(),
		req.GetRegion().GetName(),
		int(limit),
	)
	if err != nil {
		return nil, err
	}

	return &geo.GetCountiesInRegionResponse{
		Counties: srv.renderCounties(counties),
	}, nil
}

func (srv *serverImpl) GetCountiesSitemap(
	ctx context.Context,
	req *geo.GetCountiesSitemapRequest,
) (*geo.GetCountiesSitemapResponse, error) {
	counties, err := srv.Repository.CountyListByCountryRegion(ctx, req.GetCountry(), req.GetRegion(), 0)
	if err != nil {
		return nil, err
	}

	return &geo.GetCountiesSitemapResponse{
		Counties: srv.renderCounties(counties),
	}, nil
}

func (srv *serverImpl) GetTownsSitemap(
	ctx context.Context,
	req *geo.GetTownsSitemapRequest,
) (*geo.GetTownsSitemapResponse, error) {
	towns, err := srv.Repository.TownListByCountryRegionCounty(ctx, req.GetCountry(), req.GetRegion(), req.GetCounty())
	if err != nil {
		return nil, err
	}

	return &geo.GetTownsSitemapResponse{
		Towns: srv.renderTowns(towns),
	}, nil
}

func (srv *serverImpl) GetRegionsSitemap(
	ctx context.Context,
	req *geo.GetRegionsSitemapRequest,
) (*geo.GetRegionsSitemapResponse, error) {
	regions, err := srv.Repository.RegionListByCountry(ctx, req.GetCountry())
	if err != nil {
		return nil, err
	}

	return &geo.GetRegionsSitemapResponse{
		Regions: srv.renderRegions(regions),
	}, nil
}

func (srv *serverImpl) GenerateSitemaps(
	ctx context.Context,
	req *geo.GenerateSitemapsRequest,
) (*geo.GenerateSitemapsResponse, error) {
	err := srv.SitemapGenerator.Generate(ctx, srv.Config.SitemapCountries)
	if err != nil {
		return nil, err
	}

	return &geo.GenerateSitemapsResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) ClearSitemapsCache(
	ctx context.Context,
	req *geo.ClearSitemapsCacheRequest,
) (*geo.ClearSitemapsCacheResponse, error) {
	err := srv.SitemapGenerator.Clear(ctx)
	if err != nil {
		return nil, err
	}

	return &geo.ClearSitemapsCacheResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) SitemapDownload(
	req *geo.SitemapDownloadRequest,
	resp geo.GeoService_SitemapDownloadServer,
) error {
	chunk := make([]byte, 2048)

	hash, size, mime, chunker, err := srv.SitemapGenerator.Open(resp.Context(), req.GetFileName(), chunk)
	if err != nil {
		return err
	}

	err = resp.Send(&geo.FileEvent{
		Init: &geo.FileInit{
			Name:   req.GetFileName(),
			Mime:   mime,
			Size:   size,
			Sha256: hash,
		},
	})
	if err != nil {
		return err
	}

	var eof bool

	for !eof {
		var bs []byte

		bs, err = chunker.Next()
		if errors.Is(err, io.EOF) {
			eof = true
		} else if err != nil {
			return err
		}

		err = resp.Send(&geo.FileEvent{
			Data: &geo.FileData{
				Data: bs,
			},
		})
		if err != nil {
			return err
		}
	}

	return nil
}

func (srv *serverImpl) ListTimezones(
	ctx context.Context,
	req *geo.TimezonesRequest,
) (*geo.TimezonesResponse, error) {
	tzResp := make([]*geo.TimezonesResponseData, 0)

	for _, t := range req.GetTimezonesRequestData() {
		zone, err := tz.GetZone(tz.Point{
			Lon: t.Longitude,
			Lat: t.Latitude,
		})
		if err != nil {
			return nil, err
		}

		tzResp = append(tzResp, &geo.TimezonesResponseData{
			Id:       t.Id,
			Timezone: zone[0],
		})
	}

	return &geo.TimezonesResponse{
		TimezonesResponseData: tzResp,
	}, nil
}
