package sitemap

import (
	"bytes"
	"compress/gzip"
	"context"
	"crypto/sha256"
	"encoding/csv"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"

	"github.com/snabb/sitemap"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	"gitlab.mashplace.com/mashplace/backend/geo/docs"
	grabdata "gitlab.mashplace.com/mashplace/backend/grabdata/api"
)

var errNotExists = grpcerr.New(grpcerr.NotFound, "file does not exist")

type generatorImpl struct {
	config Config
}

type wordpressPagesAPIResponse struct {
	ACF json.RawMessage `json:"acf"`
}

type wordpressArticlesAPIResponse struct {
	ID         int32           `json:"id"`
	Slug       string          `json:"slug"`
	Categories json.RawMessage `json:"categories"`
}

type wordpressPagesACFDetails struct {
	PublicURL string `json:"publicurl"`
}

func (gen *generatorImpl) getPublicPagesLinks() ([]string, error) {
	resp, err := gen.config.HTTPClient.Get(gen.config.PublicPagesURL)
	if err != nil {
		return nil, err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	if resp.StatusCode != 200 {
		return nil, grpcerr.New(grpcerr.ProviderError, "provider return error code "+resp.Status)
	}

	var data []wordpressPagesAPIResponse

	err = json.NewDecoder(resp.Body).Decode(&data)
	if err != nil {
		return nil, err
	}

	var links []string

	for _, d := range data {
		if len(d.ACF) == 0 || d.ACF[0] != '{' {
			continue
		}

		acf := &wordpressPagesACFDetails{}

		err = json.Unmarshal(d.ACF, acf)
		if err != nil {
			continue
		}

		if acf.PublicURL == "" {
			continue
		}

		links = append(links, acf.PublicURL)
	}

	return links, nil
}

func (gen *generatorImpl) getArticlesPagesLinks() ([]string, error) {
	request, err := http.NewRequest(http.MethodGet, gen.config.ArticlesPagesURL, nil)
	if err != nil {
		return nil, err
	}

	offset := 0

	var links []string

	q := request.URL.Query()
	q.Add("per_page", "100")
	q.Add("offset", strconv.Itoa(offset))

	for {
		q.Set("offset", strconv.Itoa(offset))
		request.URL.RawQuery = q.Encode()

		var resp *http.Response

		resp, err = gen.config.HTTPClient.Do(request)
		if err != nil {
			return nil, err
		}

		if resp.StatusCode != 200 {
			_ = resp.Body.Close()
			return nil, grpcerr.New(grpcerr.ProviderError, "provider return error code "+resp.Status)
		}

		var data []wordpressArticlesAPIResponse

		err = json.NewDecoder(resp.Body).Decode(&data)
		if err != nil {
			_ = resp.Body.Close()
			return nil, err
		}

		for _, d := range data {
			if len(d.Categories) == 0 {
				continue
			}

			var arrCategories []int

			err = json.Unmarshal(d.Categories, &arrCategories)
			if err != nil {
				continue
			}

			if len(arrCategories) == 0 {
				continue
			}

			links = append(links, getArticlesPageLink(arrCategories, gen.config.SitemapURL, d.Slug, d.ID))
		}

		if resp.Header.Get("X-WP-Total") == "0" || len(data) == 0 {
			_ = resp.Body.Close()
			break
		}

		offset += 100
		_ = resp.Body.Close()
	}

	return links, err
}

func getArticlesPageLink(categories []int, host, slug string, id int32) (link string) {
	found := false

	for _, c := range categories {
		switch c {
		case 127:
			link = fmt.Sprintf("%s/guides/%s-%d", host, slug, id)
			found = true
		case 108:
			link = fmt.Sprintf("%s/area-guides/%s-%d", host, slug, id)
			found = true
		}

		if found {
			break
		}
	}

	if !found {
		link = fmt.Sprintf("%s/blog/%s-%d", host, slug, id)
	}

	return
}

func (gen *generatorImpl) writeSitemap(
	smi *sitemap.Sitemap,
	file string,
	idx, idxgz *sitemap.SitemapIndex,
) error {
	base := file + ".xml"

	name := path.Join(gen.config.Storage, base)
	gzname := name + ".gz"

	idx.Add(&sitemap.URL{Loc: gen.config.SitemapURL + "/sitemaps/" + base})
	idxgz.Add(&sitemap.URL{Loc: gen.config.SitemapURL + "/sitemaps/" + base + ".gz"})

	f, err := os.Create(name)
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	_, err = smi.WriteTo(f)
	if err != nil {
		return err
	}

	fgz, err := os.Create(gzname)
	if err != nil {
		return err
	}

	defer func() {
		_ = fgz.Close()
	}()

	wgz := gzip.NewWriter(fgz)

	defer func() {
		_ = wgz.Close()
	}()

	_, err = smi.WriteTo(wgz)
	if err != nil {
		return err
	}

	return nil
}

func (gen *generatorImpl) generateSitemapArticles(idx, idxgz *sitemap.SitemapIndex) error {
	smi := sitemap.New()

	articlesLinks, err := gen.getArticlesPagesLinks()
	if err != nil {
		return err
	}

	for _, a := range articlesLinks {
		smi.Add(&sitemap.URL{Loc: a})
	}

	return gen.writeSitemap(smi, "sitemaparticles", idx, idxgz)
}

func (gen *generatorImpl) generateSitemapGeneral(idx, idxgz *sitemap.SitemapIndex) error {
	smi := sitemap.New()

	generalLinks, err := gen.getPublicPagesLinks()
	if err != nil {
		return err
	}

	mp, err := getMapNoIndexUrls()
	if err != nil {
		return err
	}

	for _, l := range generalLinks {
		if !strings.HasPrefix(l, "/") {
			l = "/" + l
		}

		_, ok := mp[l]
		if ok {
			continue
		}

		smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + l})
	}

	return gen.writeSitemap(smi, "sitemapgeneral", idx, idxgz)
}

func (gen *generatorImpl) generateSitemapCountry(
	country *Country,
	idx, idxgz *sitemap.SitemapIndex,
) error {
	smi := sitemap.New()

	smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + "/to-rent"})

	return gen.writeSitemap(smi, country.File, idx, idxgz)
}

func getMapNoIndexUrls() (map[string]bool, error) {
	bs, err := docs.FS.ReadFile("no_index_urls.csv")
	if err != nil {
		return nil, err
	}

	r := csv.NewReader(bytes.NewReader(bs))

	records, err := r.ReadAll()
	if err != nil {
		return nil, err
	}

	mp := make(map[string]bool)

	for i, recs := range records {
		if i == 0 {
			continue
		}

		mp[recs[0]] = true
	}

	return mp, nil
}

func (gen *generatorImpl) generateSitemapLettingAgents(idx, idxgz *sitemap.SitemapIndex) error {
	smi := sitemap.New()
	smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + "/letting-agents"})

	return gen.writeSitemap(smi, "sitemap_lettingagents", idx, idxgz)
}

func (gen *generatorImpl) generateSitemapCountryLettingAgents(
	ctx context.Context, country *Country, idx, idxgz *sitemap.SitemapIndex,
) error {
	towns, err := gen.config.Repo.TownListByCountry(ctx, country.Name)
	if err != nil {
		return err
	}

	locationToAgentsCount, err := gen.getAgentsCounters(ctx, country.Name)
	if err != nil {
		return err
	}

	smi := sitemap.New()
	escp := func(s string) string {
		return strings.ReplaceAll(regexp.MustCompile(`[\s/]`).ReplaceAllString(strings.ToLower(s), "-"), "&", "%26")
	}

	var uniqueLocations = make(map[string]struct{})

	for _, t := range towns {
		_, ok := uniqueLocations[t.Country]
		if count := locationToAgentsCount[t.Country]; count > 1 && !ok {
			endp := fmt.Sprintf("/letting-agents/%s", escp(t.Country))

			smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + endp})

			uniqueLocations[t.Country] = struct{}{}
		}

		_, ok = uniqueLocations[t.Country+t.Region]
		if count := locationToAgentsCount[t.Country+t.Region]; count > 1 && !ok {
			endp := fmt.Sprintf("/letting-agents/%s/%s", escp(t.Country), escp(t.Region))

			smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + endp})

			uniqueLocations[t.Country+t.Region] = struct{}{}
		}

		_, ok = uniqueLocations[t.Country+t.Region+t.County]
		if count := locationToAgentsCount[t.Country+t.Region+t.County]; count > 1 && !ok {
			endp := fmt.Sprintf("/letting-agents/%s/%s/%s", escp(t.Country), escp(t.Region), escp(t.County))

			smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + endp})

			uniqueLocations[t.Country+t.Region+t.County] = struct{}{}
		}

		_, ok = uniqueLocations[t.Country+t.Region+t.County+t.Name]
		if count := locationToAgentsCount[t.Country+t.Region+t.County+t.Name]; count < 2 || ok {
			continue
		}

		endp := fmt.Sprintf("/letting-agents/%s/%s/%s/%s", escp(t.Country), escp(t.Region), escp(t.County), escp(t.Name))

		smi.Add(&sitemap.URL{Loc: gen.config.SitemapURL + endp})

		uniqueLocations[t.Country+t.Region+t.County+t.Name] = struct{}{}
	}

	return gen.writeSitemap(smi, country.File+"_lettingagents", idx, idxgz)
}

func (gen *generatorImpl) getAgentsCounters(ctx context.Context, country string) (map[string]int32, error) {
	if country == "" {
		return nil, nil
	}

	agentsCounters, err := gen.config.GrabdataCli.GetAgentsCountByLocations(
		ctx,
		&grabdata.GetAgentsCountByLocationsRequest{
			Location: &grabdata.Location{
				Country: country,
			},
		})
	if err != nil {
		return nil, err
	}

	var result = make(map[string]int32)

	for _, agentsCounter := range agentsCounters.GetAgentsCount() {
		countryKey := agentsCounter.GetLocation().GetCountry()
		countryRegionKey := countryKey + agentsCounter.GetLocation().GetRegion()
		countryRegionCountyKey := countryRegionKey + agentsCounter.GetLocation().GetCounty()
		countryRegionCountyTownKey := countryRegionCountyKey + agentsCounter.GetLocation().GetTown()

		result[countryKey] += agentsCounter.GetTotal()

		result[countryRegionKey] += agentsCounter.GetTotal()

		result[countryRegionCountyKey] += agentsCounter.GetTotal()

		result[countryRegionCountyTownKey] += agentsCounter.GetTotal()
	}

	return result, nil
}

func (gen *generatorImpl) Generate(ctx context.Context, countries []*Country) error {
	if err := os.MkdirAll(gen.config.Storage, os.FileMode(0777)); err != nil {
		return err
	}

	idx, idxgz := sitemap.NewSitemapIndex(), sitemap.NewSitemapIndex()

	if err := gen.generateSitemapGeneral(idx, idxgz); err != nil {
		return err
	}

	if err := gen.generateSitemapArticles(idx, idxgz); err != nil {
		return err
	}

	if err := gen.generateSitemapLettingAgents(idx, idxgz); err != nil {
		return err
	}

	for _, c := range countries {
		if err := gen.generateSitemapCountry(c, idx, idxgz); err != nil {
			return err
		}

		if err := gen.generateSitemapCountryLettingAgents(ctx, c, idx, idxgz); err != nil {
			return err
		}
	}

	f, err := os.Create(path.Join(gen.config.Storage, "sitemapindex.xml"))
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	_, err = idx.WriteTo(f)
	if err != nil {
		return err
	}

	fgz, err := os.Create(path.Join(gen.config.Storage, "sitemapindexgzip.xml"))
	if err != nil {
		return err
	}

	defer func() {
		_ = fgz.Close()
	}()

	_, err = idxgz.WriteTo(fgz)
	if err != nil {
		return err
	}

	return nil
}

func (gen *generatorImpl) Clear(ctx context.Context) error {
	return os.RemoveAll(gen.config.Storage)
}

type readSeekCloser interface {
	io.ReadSeeker
	io.Closer
}

type chunkerImpl struct {
	r   readSeekCloser
	buf []byte
}

func (c *chunkerImpl) Next() ([]byte, error) {
	n, err := c.r.Read(c.buf)

	return c.buf[:n], err
}

func (c *chunkerImpl) Close() error {
	return c.r.Close()
}

func (gen *generatorImpl) Open(
	ctx context.Context,
	file string,
	chunk []byte,
) (hash []byte, size int64, mime string, chunker Chunker, err error) {
	if strings.HasSuffix(file, ".gz") {
		mime = "application/x-gzip"
	} else {
		mime = "application/xml"
	}

	defer func() {
		if os.IsNotExist(err) {
			err = errNotExists
		}
	}()

	f, err := os.Open(path.Join(gen.config.Storage, file))
	if err != nil {
		return
	}

	h := sha256.New()

	size, err = io.Copy(h, f)
	if err != nil {
		return
	}

	hash = h.Sum(nil)

	_, err = f.Seek(0, io.SeekStart)
	if err != nil {
		return
	}

	chunker = &chunkerImpl{
		r:   f,
		buf: chunk,
	}

	return
}
