// Package mapbox provides API client to mapbox map provider
package mapbox

import (
	"context"
	"net/http"
	"time"
)

// default constants
const (
	DefaultGeocodingURL = "https://api.mapbox.com/geocoding/v5/mapbox.places"
	DefaultIsochroneURL = "https://api.mapbox.com/isochrone/v1/mapbox"
)

// DefaultPOITypes default points of interest categories
var DefaultPOITypes = []string{
	"accounting",
	"airport",
	"amusement_park",
	"aquarium",
	"art_gallery",
	"atm",
	"bakery",
	"bank",
	"bar",
	"beauty_salon",
	"bicycle_store",
	"book_store",
	"bowling_alley",
	"bus_station",
	"cafe",
	"campground",
	"car_dealer",
	"car_rental",
	"car_repair",
	"car_wash",
	"casino",
	"cemetery",
	"church",
	"city_hall",
	"clothing_store",
	"convenience_store",
	"courthouse",
	"dentist",
	"department_store",
	"doctor",
	"electrician",
	"electronics_store",
	"embassy",
	"fire_station",
	"florist",
	"funeral_home",
	"furniture_store",
	"gas_station",
	"gym",
	"hair_care",
	"hardware_store",
	"hindu_temple",
	"home_goods_store",
	"hospital",
	"insurance_agency",
	"jewelry_store",
	"laundry",
	"lawyer",
	"library",
	"liquor_store",
	"local_government_office",
	"locksmith",
	"lodging",
	"meal_delivery",
	"meal_takeaway",
	"mosque",
	"movie_rental",
	"movie_theater",
	"moving_company",
	"museum",
	"night_club",
	"painter",
	"park",
	"parking",
	"pet_store",
	"pharmacy",
	"physiotherapist",
	"plumber",
	"police",
	"post_office",
	"real_estate_agency",
	"restaurant",
	"roofing_contractor",
	"rv_park",
	"school",
	"shoe_store",
	"shopping_mall",
	"spa",
	"stadium",
	"storage",
	"store",
	"subway_station",
	"supermarket",
	"synagogue",
	"taxi_stand",
	"train_station",
	"transit_station",
	"travel_agency",
	"veterinary_care",
	"zoo",
	"bank",
	"garden",
	"cafe",
	"coffee",
	"tea",
	"tea house",
	"nursery",
	"bar",
	"garden",
	"post",
}

// Config for mapbox API client
type Config struct {
	GeocodingURL       string
	IsochroneURL       string
	Key                string
	POITypes           []string
	MaxContours        int
	MaxIsochronePeriod time.Duration
	HTTPClient         *http.Client
}

// NewClient returns new client instance
func NewClient(config *Config) (Client, error) {
	return &clientImpl{
		config: *config,
	}, nil
}

// Client for mapbox API
type Client interface {
	GetPOI(
		ctx context.Context,
		location *Location,
		tags []string,
		radiusInMeters int32,
	) ([]*POI, error)
	GetIsochrone(
		ctx context.Context,
		location *Location,
		periods []time.Duration,
		movingType string,
		denoise float64,
	) ([]*Isochrone, error)
}

// Location for POI
type Location struct {
	Longitude float64
	Latitude  float64
}

// POI point of interest
type POI struct {
	Type     string
	Name     string
	MakiType string
	Location Location
}

// Isochrone time-bound collection of points
type Isochrone struct {
	Time     int32
	Contours []Contour
}

// Contour collection of points
type Contour []*Location
