package mapbox

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"math"
	"net/http"
	"sort"
	"strconv"
	"strings"
	"time"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
)

type clientImpl struct {
	config Config
}

type responsePOI struct {
	data  []*POI
	error error
}

func (client *clientImpl) httpRequest(request *http.Request) ([]byte, error) {
	resp, err := client.config.HTTPClient.Do(request)
	if err != nil {
		return nil, err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	if resp.StatusCode/100 != 2 {
		return nil, grpcerr.New(grpcerr.ProviderError, "provider return unexpected code "+resp.Status)
	}

	return io.ReadAll(resp.Body)
}

func degreesToRadians(degrees float64) float64 {
	return degrees * math.Pi / 180.0
}

func getDistance(lon1 float64, lat1 float64, lon2 float64, lat2 float64) float64 {
	earthRadiusMeters := 6371000.0
	dLat := degreesToRadians(lat2 - lat1)
	dLon := degreesToRadians(lon2 - lon1)
	lat1 = degreesToRadians(lat1)
	lat2 = degreesToRadians(lat2)
	a := math.Sin(dLat/2)*math.Sin(dLat/2) + math.Sin(dLon/2)*math.Sin(dLon/2)*math.Cos(lat1)*math.Cos(lat2)
	c := 2 * math.Atan2(math.Sqrt(a), math.Sqrt(1-a))

	return earthRadiusMeters * c
}

func (client *clientImpl) getPOITag(
	ctx context.Context,
	location *Location,
	tag string,
	radiusInMeters int32,
	ch chan<- *responsePOI,
) {
	var (
		res []*POI
		err error
	)

	defer func() {
		if err != nil {
			err = fmt.Errorf("%s: %s", tag, err.Error())
		}

		ch <- &responsePOI{
			data:  res,
			error: err,
		}
	}()

	addr := fmt.Sprintf("%s/%s.json", strings.TrimRight(client.config.GeocodingURL, "/"), tag)

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, addr, nil)
	if err != nil {
		return
	}

	q := req.URL.Query()

	q.Add("limit", "10")
	q.Add("types", "poi")
	q.Add("country", "gb")
	q.Add("autocomplete", "true")
	q.Add("proximity", fmt.Sprintf("%f,%f", location.Longitude, location.Latitude))
	q.Add("access_token", client.config.Key)

	req.URL.RawQuery = q.Encode()

	bs, err := client.httpRequest(req)
	if err != nil {
		return
	}

	var data struct {
		Features []struct {
			Properties struct {
				Maki string `json:"maki"`
			} `json:"properties"`
			PlaceName string    `json:"place_name"`
			Center    []float64 `json:"center"`
		} `json:"features"`
	}

	err = json.Unmarshal(bs, &data)
	if err != nil {
		return
	}

	for _, f := range data.Features {
		if radiusInMeters > 0 &&
			getDistance(location.Longitude, location.Latitude, f.Center[0], f.Center[1]) > float64(radiusInMeters) {
			continue
		}

		res = append(res, &POI{
			Type:     tag,
			Name:     f.PlaceName,
			MakiType: f.Properties.Maki,
			Location: Location{
				Longitude: f.Center[0],
				Latitude:  f.Center[1],
			},
		})
	}
}

func (client *clientImpl) GetPOI(
	ctx context.Context,
	location *Location,
	tags []string,
	radiusInMeters int32,
) (res []*POI, err error) {
	var errs []string

	ch := make(chan *responsePOI)

	var respcount int

	for _, tag := range tags {
		respcount++

		go client.getPOITag(ctx, location, tag, radiusInMeters, ch)
	}

	for v := range ch {
		respcount--

		if v.error != nil {
			errs = append(errs, v.error.Error())
		}

		res = append(res, v.data...)

		if respcount == 0 {
			break
		}
	}

	close(ch)

	if errs != nil {
		err = errors.New(strings.Join(errs, ", "))
	}

	return
}

func uniqueDurations(t []time.Duration) (res []time.Duration) {
	m := make(map[time.Duration]bool)

	for _, i := range t {
		m[i] = true
	}

	for i := range m {
		res = append(res, i)
	}

	return
}

type responseIsochrone struct {
	data  []*Isochrone
	error error
}

func (client *clientImpl) getIsochronePeriods(
	ctx context.Context,
	location *Location,
	periods []time.Duration,
	movingType string,
	denoise float64,
	ch chan<- *responseIsochrone,
) {
	var (
		res []*Isochrone
		err error
	)

	defer func() {
		if err != nil {
			err = fmt.Errorf("%v: %s", periods, err.Error())
		}

		ch <- &responseIsochrone{
			data:  res,
			error: err,
		}
	}()

	addr := fmt.Sprintf(
		"%s/%s/%f,%f",
		strings.TrimRight(client.config.IsochroneURL, "/"),
		movingType,
		location.Longitude,
		location.Latitude,
	)

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, addr, nil)
	if err != nil {
		return
	}

	var times []string

	for _, p := range periods {
		times = append(times, strconv.FormatInt(int64(p.Minutes()), 10))
	}

	q := req.URL.Query()
	q.Add("contours_minutes", strings.Join(times, ","))
	q.Add("polygons", "true")
	q.Add("denoise", fmt.Sprintf("%f", denoise))
	q.Add("access_token", client.config.Key)

	req.URL.RawQuery = q.Encode()

	bs, err := client.httpRequest(req)
	if err != nil {
		return
	}

	var data struct {
		Features []struct {
			Property struct {
				Contour int32 `json:"contour"`
			} `json:"properties"`
			Geometry struct {
				Coordinates [][][]float64 `json:"coordinates"`
			} `json:"geometry"`
		} `json:"features"`
	}

	err = json.Unmarshal(bs, &data)
	if err != nil {
		return
	}

	for _, feature := range data.Features {
		var contours []Contour

		for _, coordinates := range feature.Geometry.Coordinates {
			var locations []*Location

			for _, loc := range coordinates {
				locations = append(locations, &Location{Longitude: loc[0], Latitude: loc[1]})
			}

			contours = append(contours, locations)
		}

		res = append(res, &Isochrone{Time: feature.Property.Contour, Contours: contours})
	}
}

func (client *clientImpl) GetIsochrone(
	ctx context.Context,
	location *Location,
	periods []time.Duration,
	movingType string,
	denoise float64,
) (res []*Isochrone, err error) {
	if len(periods) == 0 {
		return
	}

	var errs []string

	periods = uniqueDurations(periods)

	sort.Slice(periods, func(i, j int) bool {
		return periods[i] < periods[j]
	})

	if periods[len(periods)-1] > client.config.MaxIsochronePeriod {
		return nil, grpcerr.Newf(
			grpcerr.InvalidData,
			"the maximum contour time that can be specified is %s",
			client.config.MaxIsochronePeriod.String(),
		)
	}

	ch := make(chan *responseIsochrone)

	var respcount int

	for i := 0; i < len(periods); i += client.config.MaxContours {
		max := i + client.config.MaxContours
		if max > len(periods) {
			max = len(periods)
		}

		respcount++

		go client.getIsochronePeriods(ctx, location, periods[i:max], movingType, denoise, ch)
	}

	for v := range ch {
		respcount--

		if v.error != nil {
			errs = append(errs, v.error.Error())
		}

		res = append(res, v.data...)

		if respcount == 0 {
			break
		}
	}

	close(ch)

	if errs != nil {
		err = errors.New(strings.Join(errs, ", "))
	}

	return
}
